@@ -3,6 +3,7 @@
 # deepzoom_multiserver - Example web application for viewing multiple slides
 #
 # Copyright (c) 2010-2015 Carnegie Mellon University
+# Copyright (c) 2021-2022 Benjamin Gilbert
 #
 # This library is free software; you can redistribute it and/or modify it
 # under the terms of version 2.1 of the GNU Lesser General Public License
@@ -19,17 +20,36 @@
 #
 
 from collections import OrderedDict
-from flask import Flask, abort, make_response, render_template, url_for
 from io import BytesIO
-import openslide
-from openslide import OpenSlide, OpenSlideError
-from openslide.deepzoom import DeepZoomGenerator
-import os
 from optparse import OptionParser
+import os
 from threading import Lock
 
+from flask import Flask, abort, make_response, render_template, url_for
+
+if os.name == 'nt':
+    _dll_path = os.getenv('OPENSLIDE_PATH')
+    if _dll_path is not None:
+        if hasattr(os, 'add_dll_directory'):
+            # Python >= 3.8
+            with os.add_dll_directory(_dll_path):
+                import openslide
+        else:
+            # Python < 3.8
+            _orig_path = os.environ.get('PATH', '')
+            os.environ['PATH'] = _orig_path + ';' + _dll_path
+            import openslide
+
+            os.environ['PATH'] = _orig_path
+else:
+    import openslide
+
+from openslide import OpenSlide, OpenSlideCache, OpenSlideError, OpenSlideVersionError
+from openslide.deepzoom import DeepZoomGenerator
+
 SLIDE_DIR = '.'
 SLIDE_CACHE_SIZE = 10
+SLIDE_TILE_CACHE_MB = 128
 DEEPZOOM_FORMAT = 'jpeg'
 DEEPZOOM_TILE_SIZE = 254
 DEEPZOOM_OVERLAP = 1
@@ -41,18 +61,17 @@ app.config.from_object(__name__)
 app.config.from_envvar('DEEPZOOM_MULTISERVER_SETTINGS', silent=True)
 
 
-class PILBytesIO(BytesIO):
-    def fileno(self):
-        '''Classic PIL doesn't understand io.UnsupportedOperation.'''
-        raise AttributeError('Not supported')
-
-
-class _SlideCache(object):
-    def __init__(self, cache_size, dz_opts):
+class _SlideCache:
+    def __init__(self, cache_size, tile_cache_mb, dz_opts):
         self.cache_size = cache_size
         self.dz_opts = dz_opts
         self._lock = Lock()
         self._cache = OrderedDict()
+        # Share a single tile cache among all slide handles, if supported
+        try:
+            self._tile_cache = OpenSlideCache(tile_cache_mb * 1024 * 1024)
+        except OpenSlideVersionError:
+            self._tile_cache = None
 
     def get(self, path):
         with self._lock:
@@ -63,6 +82,8 @@ class _SlideCache(object):
                 return slide
 
         osr = OpenSlide(path)
+        if self._tile_cache is not None:
+            osr.set_cache(self._tile_cache)
         slide = DeepZoomGenerator(osr, **self.dz_opts)
         try:
             mpp_x = osr.properties[openslide.PROPERTY_NAME_MPP_X]
@@ -79,7 +100,7 @@ class _SlideCache(object):
         return slide
 
 
-class _Directory(object):
+class _Directory:
     def __init__(self, basedir, relpath=''):
         self.name = os.path.basename(relpath)
         self.children = []
@@ -94,7 +115,7 @@ class _Directory(object):
                 self.children.append(_SlideFile(cur_relpath))
 
 
-class _SlideFile(object):
+class _SlideFile:
     def __init__(self, relpath):
         self.name = os.path.basename(relpath)
         self.url_path = relpath
@@ -108,8 +129,10 @@ def _setup():
         'DEEPZOOM_OVERLAP': 'overlap',
         'DEEPZOOM_LIMIT_BOUNDS': 'limit_bounds',
     }
-    opts = dict((v, app.config[k]) for k, v in config_map.items())
-    app.cache = _SlideCache(app.config['SLIDE_CACHE_SIZE'], opts)
+    opts = {v: app.config[k] for k, v in config_map.items()}
+    app.cache = _SlideCache(
+        app.config['SLIDE_CACHE_SIZE'], app.config['SLIDE_TILE_CACHE_MB'], opts
+    )
 
 
 def _get_slide(path):
@@ -136,8 +159,12 @@ def index():
 def slide(path):
     slide = _get_slide(path)
     slide_url = url_for('dzi', path=path)
-    return render_template('slide-fullpage.html', slide_url=slide_url,
-            slide_filename=slide.filename, slide_mpp=slide.mpp)
+    return render_template(
+        'slide-fullpage.html',
+        slide_url=slide_url,
+        slide_filename=slide.filename,
+        slide_mpp=slide.mpp,
+    )
 
 
 @app.route('/<path:path>.dzi')
@@ -161,7 +188,7 @@ def tile(path, level, col, row, format):
     except ValueError:
         # Invalid level or coordinates
         abort(404)
-    buf = PILBytesIO()
+    buf = BytesIO()
     tile.save(buf, format, quality=app.config['DEEPZOOM_TILE_QUALITY'])
     resp = make_response(buf.getvalue())
     resp.mimetype = 'image/%s' % format
@@ -170,31 +197,72 @@ def tile(path, level, col, row, format):
 
 if __name__ == '__main__':
     parser = OptionParser(usage='Usage: %prog [options] [slide-directory]')
-    parser.add_option('-B', '--ignore-bounds', dest='DEEPZOOM_LIMIT_BOUNDS',
-                default=True, action='store_false',
-                help='display entire scan area')
-    parser.add_option('-c', '--config', metavar='FILE', dest='config',
-                help='config file')
-    parser.add_option('-d', '--debug', dest='DEBUG', action='store_true',
-                help='run in debugging mode (insecure)')
-    parser.add_option('-e', '--overlap', metavar='PIXELS',
-                dest='DEEPZOOM_OVERLAP', type='int',
-                help='overlap of adjacent tiles [1]')
-    parser.add_option('-f', '--format', metavar='{jpeg|png}',
-                dest='DEEPZOOM_FORMAT',
-                help='image format for tiles [jpeg]')
-    parser.add_option('-l', '--listen', metavar='ADDRESS', dest='host',
-                default='127.0.0.1',
-                help='address to listen on [127.0.0.1]')
-    parser.add_option('-p', '--port', metavar='PORT', dest='port',
-                type='int', default=5000,
-                help='port to listen on [5000]')
-    parser.add_option('-Q', '--quality', metavar='QUALITY',
-                dest='DEEPZOOM_TILE_QUALITY', type='int',
-                help='JPEG compression quality [75]')
-    parser.add_option('-s', '--size', metavar='PIXELS',
-                dest='DEEPZOOM_TILE_SIZE', type='int',
-                help='tile size [254]')
+    parser.add_option(
+        '-B',
+        '--ignore-bounds',
+        dest='DEEPZOOM_LIMIT_BOUNDS',
+        default=True,
+        action='store_false',
+        help='display entire scan area',
+    )
+    parser.add_option(
+        '-c', '--config', metavar='FILE', dest='config', help='config file'
+    )
+    parser.add_option(
+        '-d',
+        '--debug',
+        dest='DEBUG',
+        action='store_true',
+        help='run in debugging mode (insecure)',
+    )
+    parser.add_option(
+        '-e',
+        '--overlap',
+        metavar='PIXELS',
+        dest='DEEPZOOM_OVERLAP',
+        type='int',
+        help='overlap of adjacent tiles [1]',
+    )
+    parser.add_option(
+        '-f',
+        '--format',
+        metavar='{jpeg|png}',
+        dest='DEEPZOOM_FORMAT',
+        help='image format for tiles [jpeg]',
+    )
+    parser.add_option(
+        '-l',
+        '--listen',
+        metavar='ADDRESS',
+        dest='host',
+        default='127.0.0.1',
+        help='address to listen on [127.0.0.1]',
+    )
+    parser.add_option(
+        '-p',
+        '--port',
+        metavar='PORT',
+        dest='port',
+        type='int',
+        default=5000,
+        help='port to listen on [5000]',
+    )
+    parser.add_option(
+        '-Q',
+        '--quality',
+        metavar='QUALITY',
+        dest='DEEPZOOM_TILE_QUALITY',
+        type='int',
+        help='JPEG compression quality [75]',
+    )
+    parser.add_option(
+        '-s',
+        '--size',
+        metavar='PIXELS',
+        dest='DEEPZOOM_TILE_SIZE',
+        type='int',
+        help='tile size [254]',
+    )
 
     (opts, args) = parser.parse_args()
     # Load config file if specified
