@@ -2,6 +2,7 @@
 # openslide-python - Python bindings for the OpenSlide library
 #
 # Copyright (c) 2010-2014 Carnegie Mellon University
+# Copyright (c) 2021      Benjamin Gilbert
 #
 # This library is free software; you can redistribute it and/or modify it
 # under the terms of version 2.1 of the GNU Lesser General Public License
@@ -22,37 +23,36 @@
 This package provides Python bindings for the OpenSlide library.
 """
 
-from __future__ import division, print_function
-from PIL import Image
+from collections.abc import Mapping
 
-try:
-    # Python 3.3+
-    from collections.abc import Mapping
-except ImportError:
-    # Python 2
-    from collections import Mapping
+from PIL import Image
 
 from openslide import lowlevel
 
 # For the benefit of library users
-from openslide.lowlevel import OpenSlideError, OpenSlideUnsupportedFormatError
-from openslide._version import __version__
+from openslide._version import __version__  # noqa: F401  module-imported-but-unused
+from openslide.lowlevel import (  # noqa: F401  module-imported-but-unused
+    OpenSlideError,
+    OpenSlideUnsupportedFormatError,
+    OpenSlideVersionError,
+)
 
 __library_version__ = lowlevel.get_version()
 
-PROPERTY_NAME_COMMENT          = u'openslide.comment'
-PROPERTY_NAME_VENDOR           = u'openslide.vendor'
-PROPERTY_NAME_QUICKHASH1       = u'openslide.quickhash-1'
-PROPERTY_NAME_BACKGROUND_COLOR = u'openslide.background-color'
-PROPERTY_NAME_OBJECTIVE_POWER  = u'openslide.objective-power'
-PROPERTY_NAME_MPP_X            = u'openslide.mpp-x'
-PROPERTY_NAME_MPP_Y            = u'openslide.mpp-y'
-PROPERTY_NAME_BOUNDS_X         = u'openslide.bounds-x'
-PROPERTY_NAME_BOUNDS_Y         = u'openslide.bounds-y'
-PROPERTY_NAME_BOUNDS_WIDTH     = u'openslide.bounds-width'
-PROPERTY_NAME_BOUNDS_HEIGHT    = u'openslide.bounds-height'
+PROPERTY_NAME_COMMENT = 'openslide.comment'
+PROPERTY_NAME_VENDOR = 'openslide.vendor'
+PROPERTY_NAME_QUICKHASH1 = 'openslide.quickhash-1'
+PROPERTY_NAME_BACKGROUND_COLOR = 'openslide.background-color'
+PROPERTY_NAME_OBJECTIVE_POWER = 'openslide.objective-power'
+PROPERTY_NAME_MPP_X = 'openslide.mpp-x'
+PROPERTY_NAME_MPP_Y = 'openslide.mpp-y'
+PROPERTY_NAME_BOUNDS_X = 'openslide.bounds-x'
+PROPERTY_NAME_BOUNDS_Y = 'openslide.bounds-y'
+PROPERTY_NAME_BOUNDS_WIDTH = 'openslide.bounds-width'
+PROPERTY_NAME_BOUNDS_HEIGHT = 'openslide.bounds-height'
 
-class AbstractSlide(object):
+
+class AbstractSlide:
     """The base class of a slide object."""
 
     def __enter__(self):
@@ -124,20 +124,26 @@ class AbstractSlide(object):
         size:     (width, height) tuple giving the region size."""
         raise NotImplementedError
 
+    def set_cache(self, cache):
+        """Use the specified cache to store recently decoded slide tiles.
+
+        cache: an OpenSlideCache object."""
+        raise NotImplementedError
+
     def get_thumbnail(self, size):
         """Return a PIL.Image containing an RGB thumbnail of the image.
 
         size:     the maximum size of the thumbnail."""
-        downsample = max(*[dim / thumb for dim, thumb in
-                zip(self.dimensions, size)])
+        downsample = max(*(dim / thumb for dim, thumb in zip(self.dimensions, size)))
         level = self.get_best_level_for_downsample(downsample)
         tile = self.read_region((0, 0), level, self.level_dimensions[level])
         # Apply on solid background
-        bg_color = '#' + self.properties.get(PROPERTY_NAME_BACKGROUND_COLOR,
-                'ffffff')
+        bg_color = '#' + self.properties.get(PROPERTY_NAME_BACKGROUND_COLOR, 'ffffff')
         thumb = Image.new('RGB', tile.size, bg_color)
         thumb.paste(tile, None, tile)
-        thumb.thumbnail(size, Image.ANTIALIAS)
+        # Image.Resampling added in Pillow 9.1.0
+        # Image.LANCZOS removed in Pillow 10
+        thumb.thumbnail(size, getattr(Image, 'Resampling', Image).LANCZOS)
         return thumb
 
 
@@ -157,17 +163,17 @@ class OpenSlide(AbstractSlide):
         """Open a whole-slide image."""
         AbstractSlide.__init__(self)
         self._filename = filename
-        self._osr = lowlevel.open(filename)
+        self._osr = lowlevel.open(str(filename))
 
     def __repr__(self):
-        return '%s(%r)' % (self.__class__.__name__, self._filename)
+        return f'{self.__class__.__name__}({self._filename!r})'
 
     @classmethod
     def detect_format(cls, filename):
         """Return a string describing the format vendor of the specified file.
 
         If the file format is not recognized, return None."""
-        return lowlevel.detect_vendor(filename)
+        return lowlevel.detect_vendor(str(filename))
 
     def close(self):
         """Close the OpenSlide object."""
@@ -183,16 +189,18 @@ class OpenSlide(AbstractSlide):
         """A list of (width, height) tuples, one for each level of the image.
 
         level_dimensions[n] contains the dimensions of level n."""
-        return tuple(lowlevel.get_level_dimensions(self._osr, i)
-                for i in range(self.level_count))
+        return tuple(
+            lowlevel.get_level_dimensions(self._osr, i) for i in range(self.level_count)
+        )
 
     @property
     def level_downsamples(self):
         """A list of downsampling factors for each level of the image.
 
         level_downsample[n] contains the downsample factor of level n."""
-        return tuple(lowlevel.get_level_downsample(self._osr, i)
-                for i in range(self.level_count))
+        return tuple(
+            lowlevel.get_level_downsample(self._osr, i) for i in range(self.level_count)
+        )
 
     @property
     def properties(self):
@@ -225,8 +233,21 @@ class OpenSlide(AbstractSlide):
 
         Unlike in the C interface, the image data returned by this
         function is not premultiplied."""
-        return lowlevel.read_region(self._osr, location[0], location[1],
-                level, size[0], size[1])
+        return lowlevel.read_region(
+            self._osr, location[0], location[1], level, size[0], size[1]
+        )
+
+    def set_cache(self, cache):
+        """Use the specified cache to store recently decoded slide tiles.
+
+        By default, the object has a private cache with a default size.
+
+        cache: an OpenSlideCache object."""
+        try:
+            llcache = cache._openslide_cache
+        except AttributeError:
+            raise TypeError('Not a cache object')
+        lowlevel.set_cache(self._osr, llcache)
 
 
 class _OpenSlideMap(Mapping):
@@ -234,7 +255,7 @@ class _OpenSlideMap(Mapping):
         self._osr = osr
 
     def __repr__(self):
-        return '<%s %r>' % (self.__class__.__name__, dict(self))
+        return f'<{self.__class__.__name__} {dict(self)!r}>'
 
     def __len__(self):
         return len(self._keys())
@@ -268,6 +289,23 @@ class _AssociatedImageMap(_OpenSlideMap):
         return lowlevel.read_associated_image(self._osr, key)
 
 
+class OpenSlideCache:
+    """An in-memory tile cache.
+
+    Tile caches can be attached to one or more OpenSlide objects with
+    OpenSlide.set_cache() to cache recently-decoded tiles.  By default,
+    each OpenSlide object has its own cache with a default size.
+    """
+
+    def __init__(self, capacity):
+        """Create a tile cache with the specified capacity in bytes."""
+        self._capacity = capacity
+        self._openslide_cache = lowlevel.cache_create(capacity)
+
+    def __repr__(self):
+        return f'{self.__class__.__name__}({self._capacity!r})'
+
+
 class ImageSlide(AbstractSlide):
     """A wrapper for a PIL.Image that provides the OpenSlide interface."""
 
@@ -285,7 +323,7 @@ class ImageSlide(AbstractSlide):
             self._image = Image.open(file)
 
     def __repr__(self):
-        return '%s(%r)' % (self.__class__.__name__, self._file_arg)
+        return f'{self.__class__.__name__}({self._file_arg!r})'
 
     @classmethod
     def detect_format(cls, filename):
@@ -293,21 +331,15 @@ class ImageSlide(AbstractSlide):
 
         If the file format is not recognized, return None."""
         try:
-            img = Image.open(filename)
-            format = img.format
-            if hasattr(img, 'close'):
-                # Pillow >= 2.5.0
-                img.close()
-            return format
-        except IOError:
+            with Image.open(filename) as img:
+                return img.format
+        except OSError:
             return None
 
     def close(self):
         """Close the slide object."""
         if self._close:
-            if hasattr(self._image, 'close'):
-                # Pillow >= 2.5.0
-                self._image.close()
+            self._image.close()
             self._close = False
         self._image = None
 
@@ -358,26 +390,36 @@ class ImageSlide(AbstractSlide):
         if level != 0:
             raise OpenSlideError("Invalid level")
         if ['fail' for s in size if s < 0]:
-            raise OpenSlideError("Size %s must be non-negative" % (size,))
+            raise OpenSlideError(f"Size {size} must be non-negative")
         # Any corner of the requested region may be outside the bounds of
         # the image.  Create a transparent tile of the correct size and
         # paste the valid part of the region into the correct location.
-        image_topleft = [max(0, min(l, limit - 1))
-                    for l, limit in zip(location, self._image.size)]
-        image_bottomright = [max(0, min(l + s - 1, limit - 1))
-                    for l, s, limit in zip(location, size, self._image.size)]
+        image_topleft = [
+            max(0, min(l, limit - 1)) for l, limit in zip(location, self._image.size)
+        ]
+        image_bottomright = [
+            max(0, min(l + s - 1, limit - 1))
+            for l, s, limit in zip(location, size, self._image.size)
+        ]
         tile = Image.new("RGBA", size, (0,) * 4)
-        if not ['fail' for tl, br in zip(image_topleft, image_bottomright)
-                if br - tl < 0]:  # "< 0" not a typo
+        if not [
+            'fail' for tl, br in zip(image_topleft, image_bottomright) if br - tl < 0
+        ]:  # "< 0" not a typo
             # Crop size is greater than zero in both dimensions.
             # PIL thinks the bottom right is the first *excluded* pixel
-            crop = self._image.crop(image_topleft +
-                    [d + 1 for d in image_bottomright])
-            tile_offset = tuple(il - l for il, l in
-                    zip(image_topleft, location))
+            crop = self._image.crop(image_topleft + [d + 1 for d in image_bottomright])
+            tile_offset = tuple(il - l for il, l in zip(image_topleft, location))
             tile.paste(crop, tile_offset)
         return tile
 
+    def set_cache(self, cache):
+        """Use the specified cache to store recently decoded slide tiles.
+
+        ImageSlide does not support caching, so this method does nothing.
+
+        cache: an OpenSlideCache object."""
+        pass
+
 
 def open_slide(filename):
     """Open a whole-slide or regular image.
@@ -392,6 +434,7 @@ def open_slide(filename):
 
 if __name__ == '__main__':
     import sys
+
     print("OpenSlide vendor:", OpenSlide.detect_format(sys.argv[1]))
     print("PIL format:", ImageSlide.detect_format(sys.argv[1]))
     with open_slide(sys.argv[1]) as _slide:
