@@ -1,7 +1,7 @@
 #
 # openslide-python - Python bindings for the OpenSlide library
 #
-# Copyright (c) 2016 Benjamin Gilbert
+# Copyright (c) 2016-2021 Benjamin Gilbert
 #
 # This library is free software; you can redistribute it and/or modify it
 # under the terms of version 2.1 of the GNU Lesser General Public License
@@ -17,54 +17,39 @@
 # Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 #
 
-from contextlib import contextmanager
-from openslide import ImageSlide, OpenSlideError
-from PIL import Image
 import unittest
 
-from . import file_path, image_dimensions_cannot_be_zero, skip_if
+from PIL import Image
 
-# Tests should be written to be compatible with Python 2.6 unittest.
+from openslide import ImageSlide, OpenSlideCache, OpenSlideError
 
-@contextmanager
-def image_open(*args, **kwargs):
-    img = Image.open(*args, **kwargs)
-    try:
-        yield img
-    finally:
-        if hasattr(img, 'close'):
-            # Pillow >= 2.5.0
-            img.close()
+from . import file_path, maybe_supported
 
 
 class TestImageWithoutOpening(unittest.TestCase):
     def test_detect_format(self):
-        self.assertTrue(
-                ImageSlide.detect_format(file_path('__missing_file')) is None)
-        self.assertTrue(
-                ImageSlide.detect_format(file_path('../setup.py')) is None)
-        self.assertEqual(
-                ImageSlide.detect_format(file_path('boxes.png')), 'PNG')
+        self.assertTrue(ImageSlide.detect_format(file_path('__missing_file')) is None)
+        self.assertTrue(ImageSlide.detect_format(file_path('../setup.py')) is None)
+        self.assertEqual(ImageSlide.detect_format(file_path('boxes.png')), 'PNG')
 
     def test_open(self):
-        self.assertRaises(IOError,
-                lambda: ImageSlide(file_path('__does_not_exist')))
-        self.assertRaises(IOError,
-                lambda: ImageSlide(file_path('../setup.py')))
+        self.assertRaises(OSError, lambda: ImageSlide(file_path('__does_not_exist')))
+        self.assertRaises(OSError, lambda: ImageSlide(file_path('../setup.py')))
 
     def test_open_image(self):
         # passing PIL.Image to ImageSlide
-        with image_open(file_path('boxes.png')) as img:
+        with Image.open(file_path('boxes.png')) as img:
             with ImageSlide(img) as osr:
                 self.assertEqual(osr.dimensions, (300, 250))
                 self.assertEqual(repr(osr), 'ImageSlide(%r)' % img)
 
     def test_operations_on_closed_handle(self):
-        with image_open(file_path('boxes.png')) as img:
+        with Image.open(file_path('boxes.png')) as img:
             osr = ImageSlide(img)
             osr.close()
-            self.assertRaises(AttributeError,
-                    lambda: osr.read_region((0, 0), 0, (100, 100)))
+            self.assertRaises(
+                AttributeError, lambda: osr.read_region((0, 0), 0, (100, 100))
+            )
             # If an Image is passed to the constructor, ImageSlide.close()
             # shouldn't close it
             self.assertEqual(img.getpixel((0, 0)), 3)
@@ -73,8 +58,9 @@ class TestImageWithoutOpening(unittest.TestCase):
         osr = ImageSlide(file_path('boxes.png'))
         with osr:
             pass
-        self.assertRaises(AttributeError,
-                lambda: osr.read_region((0, 0), 0, (100, 100)))
+        self.assertRaises(
+            AttributeError, lambda: osr.read_region((0, 0), 0, (100, 100))
+        )
 
 
 class TestImage(unittest.TestCase):
@@ -85,8 +71,7 @@ class TestImage(unittest.TestCase):
         self.osr.close()
 
     def test_repr(self):
-        self.assertEqual(repr(self.osr),
-                'ImageSlide(%r)' % file_path('boxes.png'))
+        self.assertEqual(repr(self.osr), 'ImageSlide(%r)' % file_path('boxes.png'))
 
     def test_metadata(self):
         self.assertEqual(self.osr.level_count, 1)
@@ -101,21 +86,27 @@ class TestImage(unittest.TestCase):
         self.assertEqual(self.osr.associated_images, {})
 
     def test_read_region(self):
-        self.assertEqual(self.osr.read_region((-10, -10), 0, (400, 400)).size,
-                (400, 400))
+        self.assertEqual(
+            self.osr.read_region((-10, -10), 0, (400, 400)).size, (400, 400)
+        )
 
-    @skip_if(image_dimensions_cannot_be_zero, 'Pillow issue #2259')
     def test_read_region_size_dimension_zero(self):
-        self.assertEqual(self.osr.read_region((0, 0), 0, (400, 0)).size,
-                (400, 0))
+        self.assertEqual(self.osr.read_region((0, 0), 0, (400, 0)).size, (400, 0))
 
     def test_read_region_bad_level(self):
-        self.assertRaises(OpenSlideError,
-                lambda: self.osr.read_region((0, 0), 1, (100, 100)))
+        self.assertRaises(
+            OpenSlideError, lambda: self.osr.read_region((0, 0), 1, (100, 100))
+        )
 
     def test_read_region_bad_size(self):
-        self.assertRaises(OpenSlideError,
-                lambda: self.osr.read_region((0, 0), 0, (400, -5)))
+        self.assertRaises(
+            OpenSlideError, lambda: self.osr.read_region((0, 0), 0, (400, -5))
+        )
 
     def test_thumbnail(self):
         self.assertEqual(self.osr.get_thumbnail((100, 100)).size, (100, 83))
+
+    @maybe_supported
+    def test_set_cache(self):
+        self.osr.set_cache(OpenSlideCache(64 << 10))
+        self.assertEqual(self.osr.read_region((0, 0), 0, (400, 400)).size, (400, 400))
